#!/usr/bin/env python
#This is public domain source code
#If you are interested in citing this work in an academic context, it was
#introduced with our forthcoming reg2XML poster:
#Andrew I. Schein, Simson L. Garfinkel, Joshua B. Gross.
#reg2xml: A Tool for Windows Registry Analysis. 2010. Under Review


import getopt, re, sys
import xml.parsers.expat as expat
from sys import argv, stdout, stderr
from os import path, environ, unlink 
if "DOMEX_HOME" not in environ : raise RuntimeError("$DOMEX_HOME is not defined") #import fiwalk
fiwalkpth=environ["DOMEX_HOME"] + "/src/fiwalk/python/"
if not path.exists(fiwalkpth) : raise RuntimeError(fiwalkpth + " not found.")
sys.path = [fiwalkpth] + sys.path
import fiwalk
E=environ.get

quiet       = False
progname    = path.basename(argv[0])
usage       = """%s pattern fiwalk_output.xml imagefile.aff out_filename\n\
Extracts files matching pattern. Defaults to exact match\n\
Writes output to standard out without optional argument\n\
In case of multiple matches, only reacts to first\n\
-r treat pattern as a regular expression (expects quotes to strip)\n\
-e use a string.endswidth() match (incompatible with -r)\n\
-i case insensitive matching\n\
-q do not report when pattern fails to match""" %progname

### SET UP COMMAND LINE ARG PARSING
try:
    opts, argv = getopt.getopt(argv[1:], "ire")
except getopt.GetoptError, err: # print help information and exit:
    print >>stderr, str(err) # will print something like "option -a not recognized"
    usage()
    sys.exit(2)
ignore_case = False #case insensitive
endsWith    = False #use endsWith match (for non-regular expression patterns)
regExp      = None  #interpret pattern as regular expression
reFlags     = 0

for o,a in opts :
    if   o in ("-i") : ignore_case=True ; reFlags=re.IGNORECASE ; 
    elif o in ("-r") : regExp=True
    elif o in ("-e") : endsWith=True
    elif o in ("-q") : quiet=True
    else : raise RuntimeException("argument mismatch")
if endsWith and regExp : raise RuntimeException("-e and -r are incompatible.")
if regExp is not None  : regExp = re.compile(argv[0].strip("\""),reFlags)
# hard_code=".*system32.config.system$"
# if hard_code!=argv[0].strip("\"") :
#     print hard_code
#     print argv[0]
#     exit(1)
if len(argv) < 3       : print >>stdout, usage ; exit(1)
if "AFFLIB_PASSPHRASE" not in environ : print >>stderr, "Warning: AFFLIB_PASSPHRASE not defined." 
VERBOSE=E("VERBOSE",None)
PATTERN=0; XML=1; IMG=2; OUT=3 
pattern   = argv[PATTERN]
xml_infile= file(argv[XML]) ; image_file=file(argv[IMG])
found_one = False #only reacts to first match.

def cb(fo) : #callback acts on each fileobject
    global found_one
    fname = fo.filename() if not ignore_case else fo.filename().lower()
    
    if not fo.allocated() : return
    if (regExp is None and fname.endswith(pattern)) or \
            (regExp is not None and regExp.match(fname)):
        if not VERBOSE is None : print "extracting"
        if fo.has_tag("inode") and fo.has_contents() and not found_one :
            found_one = True
            imgfile   = file(argv[IMG],"rb")
            outfile   = file(argv[OUT],"w")
            try :
                outfile.write(fo.contents(imagefile=image_file)) #found_one guards output file
            except ValueError, e :
                unlink(argv[OUT]) #remove incomplete file
                raise ValueError, e
            

### DRIVE OPERATIONS
if ignore_case : pattern = pattern.lower()
try :
    fileobjects = fiwalk.fiwalk_using_sax(xmlfile=xml_infile,imagefile=image_file,callback=cb)
except expat.ExpatError, e :
    print >>stderr, e
    exit(1)
if not found_one and not quiet : print >>stderr, "Pattern not found."
exit(not found_one)
