#include "config.h"

#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64
#endif

#ifndef __USE_FILE_OFFSET64
#define __USE_FILE_OFFSET64
#endif

#include "utils.h"

#include <ctype.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <inttypes.h>
#include <stdlib.h>
#include <sys/stat.h>

#ifndef HAVE_ERR
#include <stdarg.h>
void err(int eval,const char *fmt,...)
{
  va_list ap;
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
  fprintf(stderr,": %s\n",strerror(errno));
  exit(eval);
}
#endif

#ifndef HAVE_ERRX
#include <stdarg.h>
void errx(int eval,const char *fmt,...)
{
  va_list ap;
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  fprintf(stderr,"%s\n",strerror(errno));
  va_end(ap);
  exit(eval);
}
#endif

#ifndef HAVE_WARN
#include <stdarg.h>
void	warn(const char *fmt, ...)
{
    va_list args;
    va_start(args,fmt);
    vfprintf(stderr,fmt, args);
    fprintf(stderr,": %s\n",strerror(errno));
}
#endif

#ifndef HAVE_WARNX
#include <stdarg.h>
void warnx(const char *fmt,...)
{
  va_list ap;
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
}
#endif

#ifndef HAVE_ISHEXNUMBER
int ishexnumber(int c)
{
    switch(c){
    case '0':         case '1':         case '2':         case '3':         case '4':
    case '5':         case '6':         case '7':         case '8':         case '9':
    case 'A':         case 'B':         case 'C':         case 'D':         case 'E':
    case 'F':         case 'a':         case 'b':         case 'c':         case 'd':
    case 'e':         case 'f':
	return 1;
    }
    return 0;
}
#endif

int64_t atoi64(const char *buf)
{
    int64_t ret=0;
    sscanf(buf,"%"PRId64,&ret);
    return ret;
}

/** Extract a buffer...
 * @param buf - the buffer to extract;
 * @param buflen - the size of the page to extract
 * @param pos0 - the byte position of buf[0]
 */

int ends_with(const char *buf,const char *with)
{
    if(buf && with){
	size_t buflen = strlen(buf);
	size_t withlen = strlen(with);
	if(buflen>withlen && strcmp(buf+buflen-withlen,with)==0) return 1;
    }
    return 0;
}

/**
 * Prints some of a buffer, but only printable characters...
 */
void fputascii(FILE *f,unsigned char *buf,int len)
{
    for(;len>0;len--,buf++){
	if(*buf == '|' || !isprint(*buf)){
	    fputc('.',f);
	}
	else {
	    fputc(*buf,f);
	}
    }
}

/* Figure out what to do for lseek */
#if defined(HAVE_LSEEKO)
#define LSEEK lseeko
#endif

#if defined(HAVE_LSEEKO64) && !defined(LSEEK)
#define LSEEK lseeko64
#endif

#if defined(HAVE__LSEEKI64) && !defined(LSEEK)
#define LSEEK _lseeki64
#endif

#if !defined(LSEEK)
#define LSEEK lseek
#endif



/**
 * It's hard to figure out the filesize in an opearting system independent method that works with both
 * files and devices. This seems to work. It only requires a functioning seek.
 */

int64_t get_filesize(int fd)
{
    struct stat st;

    char buf[64];
    uint64_t raw_filesize = 0;
    int bits = 0;
    int i =0;

    /* We can use fstat if sizeof(st_size)==8 and st_size>0 */
    if(sizeof(st.st_size)==8 && fstat(fd,&st)==0){
	if(st.st_size>0) return st.st_size;
    }

    /* Phase 1; figure out how far we can seek... */
    for(bits=0;bits<60;bits++){
	raw_filesize = ((uint64_t)1<<bits);
	if((LSEEK(fd,raw_filesize,SEEK_SET)<0) || (read(fd,buf,1)!=1)){
	    break;
	}
    }
    if(bits==60) errx(1,"Partition detection not functional.\n");

    /* Phase 2; blank bits as necessary */
    for(i=bits;i>=0;i--){
	uint64_t test = (uint64_t)1<<i;
	uint64_t test_filesize = raw_filesize | ((uint64_t)1<<i);
	if((LSEEK(fd,test_filesize,SEEK_SET)>=0) &&
	   (read(fd,buf,1)==1)){
	    raw_filesize |= test;
	} else{
	    raw_filesize &= ~test;
	}
    }
    if(raw_filesize>0) raw_filesize+=1;	/* seems to be needed */
    return raw_filesize;
}

